/*
 * uart.h
 *
 * Created: 20.02.2019 20:50:01
 *  Author: jens
 */ 


#ifndef UART_H_
#define UART_H_

#include "machine.h"

#define XON                     17       /* XON Zeichen */
#define XOFF                    19       /* XOFF Zeichen */
#define UART_PRESCALE(baudrate, smpPerBit) ((5+ 10*F_CPU / (baudrate*smpPerBit) - 1)/10)

void uart_init(uint32_t baudrate);
void uart_putc(char c);
char uart_getc(void);
void uart_puts(char const *str);

void print_byte(char byte);
void print_word(uint16_t word);

#endif /* UART_H_ */